/*
 * sdl.cxx            
 *                           
 *This file contains the class heirachy for 1)creating windows under sdl.
 *                                          2)displaying data on those windows
 *
 * Copyright (c) 1999-2000 Indranet Technologies ltd
 * 
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.0 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
 * the License for the specific language governing rights and limitations
 * under the License.
 * 
 * The Original Code is Open H323
 * 
 * Contributor(s): Author Derek J Smithies (derek@indranet.co.nz)
 *
 * $Log: sdlvid.cxx,v $
 * Revision 1.11  2002/05/06 01:21:37  dereks
 * Improve event handling, so better reliability.
 *
 * Revision 1.10  2002/04/29 03:51:55  dereks
 * Documentation tidy up. Thanks to Walter Whitlock.
 *
 * Revision 1.9  2002/04/26 03:33:32  dereks
 * Major upgrade. All calls to SDL library are now done by one thread.
 *
 * Revision 1.8  2002/01/04 04:17:52  dereks
 * Add code from Walter Whitlock to flip video, and improve the
 * --videotest mode, which displays local video (without a call).  Many thanks.
 *
 * Revision 1.7  2001/12/13 03:01:57  dereks
 * Modify trace statement
 *
 * Revision 1.6  2001/11/20 21:03:28  dereks
 * Fix ANSI C++ complaints about pointer arithmetic with void *
 *
 * Revision 1.5  2001/11/19 01:10:24  dereks
 * Various bug fixes in sdl display of video.
 *
 * Revision 1.4  2001/10/23 02:21:39  dereks
 * Initial release CU30 video codec.
 * Add --videotest option, to display raw video, but not invoke a call.
 *
 * Revision 1.3  2001/08/02 05:13:39  dereks
 * Added PTRACE statments for debugging.
 *
 * Revision 1.2  2001/05/25 01:14:44  dereks
 * Alter SetFrameSize & OpenWindo to use unsigned variables. Change type of
 * other variables to eliminate compiler warnings.
 *
 * Revision 1.1  2001/03/07 01:47:45  dereks
 * Initial release of SDL (Simple DirectMedia Layer, a cross-platform multimedia library),
 * a video library code.
 *
 *
 *
 */

#include <ptlib.h>

#ifdef HAS_SDL
#include "sdlvid.h"
OneVideoFrame::OneVideoFrame(unsigned newWidth, unsigned newHeight, const void *_data)
{
  Initialise(newWidth, newHeight, (u_char *)_data);
}

OneVideoFrame::OneVideoFrame(unsigned newWidth, unsigned newHeight, u_char *_data)
{
  Initialise(newWidth, newHeight, _data);
}

OneVideoFrame::~OneVideoFrame()
{
  delete data;
}

void OneVideoFrame::Initialise(unsigned newWidth, 
			       unsigned newHeight, u_char *_data)
{
  width = newWidth;
  height = newHeight;
  unsigned size = (width*height *3) >> 1;
  data = new u_char[size];
  memcpy(data, _data, size);
}
//////////////////////////////////////////////////////////////


////////////////////////////////////////////////////
////////////////////////////////////////////////////
////////////////////////////////////////////////////
////////////////////////////////////////////////////
////////////////////////////////////////////////////
////////////////////////////////////////////////////
////////////////////////////////////////////////////
////////////////////////////////////////////////////

SDLVideoDevice::SDLVideoDevice(const PString & _remoteName,
			       BOOL _isEncoding, 
                               SdlDisplayThread *_sdlThread)
{
  PTRACE(3,"Generic\t SDL video constructor start.");
  
  remoteName = _remoteName; 
  isEncoding = _isEncoding;
  sdlThread  = _sdlThread;
       
  width=0;
  height=0;

  sdlThread->RequestOpenWindow(isEncoding);
}

SDLVideoDevice::~SDLVideoDevice()
{ 
  Close();
}

BOOL SDLVideoDevice::Close()
{
  sdlThread->RequestCloseWindow(isEncoding);
  return TRUE;
}

BOOL SDLVideoDevice::SetFrameSize (unsigned _width, unsigned _height)
{
  width = _width;
  height = _height;

  return TRUE;
}

BOOL SDLVideoDevice::Redraw (const void *frame)
{
  OneVideoFrame *newWindow = new OneVideoFrame(width, height, frame);

  sdlThread->AddFrame(newWindow, isEncoding);

  return TRUE;
}

BOOL  SDLVideoDevice::IsOpen()
{
  return TRUE;
}

//////////////////////////////////////////////////////////////

SdlDisplayThread::SdlDisplayThread(BOOL _videoPIP)
      : PThread(1000, NoAutoDeleteThread, LowPriority, "SDL display")
{
  for(int i = 0; i < 2; i++) {
    width[i] = 0;
    height[i] = 0;
    overlay[i] = NULL;
    displayPosn[i].x = 0;
    displayPosn[i].y = 0;
    displayPosn[i].w = 0;
    displayPosn[i].h = 0;
  }  

  oldScreenWidth = 0;
  oldScreenHeight = 0;

  closeEncWindow = FALSE;
  closeRecWindow = FALSE;
  
  screen     = NULL;
  threadOpen = TRUE;
  displayIsShutDown = FALSE;

  nextEncFrame = NULL;
  nextRcvFrame = NULL;
  videoPIP = _videoPIP;
  Resume();
}

SdlDisplayThread::~SdlDisplayThread()
{
}

BOOL SdlDisplayThread::AddFrame(OneVideoFrame *newFrame, BOOL isEncoding)
{
  accessLock.Wait();

  if (!IsOpen()) { //This frame not wanted, as the display thread is closed.
    PTRACE(2, "SDL\tDelete unwanted " << GetDirName(isEncoding) << "frame");
    delete newFrame;
    accessLock.Signal();
    return FALSE;
  }

  OneVideoFrame **frameQ;

  if (isEncoding) 
    frameQ = &nextEncFrame;
  else
    frameQ = &nextRcvFrame;

  if ((*frameQ) != NULL) {
    PTRACE(2, "SDL\tDelete unused " << GetDirName(isEncoding) << "frame");
    delete *frameQ;
  }

  *frameQ = newFrame;

  accessLock.Signal();  
  readFlag.Signal();
  return TRUE;
}

BOOL SdlDisplayThread::IsOpen()
{
  return threadOpen;
}

BOOL SdlDisplayThread::ScreenIsOpen()
{
  return (screen != NULL);
}

BOOL SdlDisplayThread::DisplayIsShutDown()
{
  return displayIsShutDown;
}

void SdlDisplayThread::Terminate()
{
  PTRACE(3, "SDL\tSdl has been told to terminate. Set flags to close");
  accessLock.Wait();
  
  if (nextEncFrame != NULL)
    delete nextEncFrame;

  if (nextRcvFrame != NULL)
    delete nextRcvFrame;

  nextEncFrame = NULL;
  nextRcvFrame = NULL;

  threadOpen = FALSE;

  accessLock.Signal();
  readFlag.Signal();
}

void SdlDisplayThread::RequestCloseWindow(BOOL isEncoding)
{
  PTRACE(3, "SDL\tRequest: Close window " << GetDirName(isEncoding) << " video");

  accessLock.Wait();

  if (isEncoding)
    closeEncWindow = TRUE;
  else
    closeRecWindow = TRUE;

  accessLock.Signal();
  readFlag.Signal();
}

void SdlDisplayThread::RequestOpenWindow(BOOL isEncoding)
{
  PTRACE(3, "SDL\tRequest: Open window " << GetDirName(isEncoding) << " video");

  accessLock.Wait();
  if (isEncoding)
    closeEncWindow = FALSE;
  else
    closeRecWindow = FALSE;

  accessLock.Signal();
}


void SdlDisplayThread::AwaitTermination()
{
  exitFlag.Wait();
}


OneVideoFrame *SdlDisplayThread::GetNextFrame(BOOL isEncoding)
{
  OneVideoFrame *res;
  accessLock.Wait();

  if (!IsOpen()) {
    accessLock.Signal();
    return FALSE;
  }

  OneVideoFrame **frameQ;

  if (isEncoding) 
    frameQ = &nextEncFrame;
  else
    frameQ = &nextRcvFrame;

  res = *frameQ;
  *frameQ = NULL;
  accessLock.Signal();  

  return res;
}

BOOL SdlDisplayThread::InitialiseSdl()
{
  if ( SDL_Init(SDL_INIT_VIDEO) < 0 ) {
    PTRACE(0,"Couldn't initialize SDL: " << SDL_GetError());
    return FALSE;      
    }

  return TRUE;
}

BOOL SdlDisplayThread::ResizeScreen(unsigned newWidth, unsigned newHeight)
{   
  if ((oldScreenWidth == newWidth) && (oldScreenHeight == newHeight)) 
    return TRUE;

  oldScreenWidth  = newWidth;
  oldScreenHeight = newHeight;
 
  PTRACE(3,"SDL Open SDL Screen of size " << newWidth << "x" << newHeight);
  
  WriteOutDisplay();

  screen = SDL_SetVideoMode(newWidth, newHeight, 0, SDL_SWSURFACE | SDL_RESIZABLE );
  
  if (!ScreenIsOpen()) {
    PTRACE(0,"Could not open screen to display window: " << SDL_GetError());
    SDL_Quit();
    return FALSE;
  }
  
  PTRACE(3,"SDL\tSuccessfully resize a SDL screen. New size= " << newWidth << "x" << newHeight);
  return TRUE;
}

void SdlDisplayThread::InitDisplayPosn()
{
  SDL_Surface * currentSurface = SDL_GetVideoSurface();
  if (currentSurface != NULL)
    InitDisplayPosn(currentSurface->w, currentSurface->h);
}

void SdlDisplayThread::InitDisplayPosn(unsigned w, unsigned h)
{  
  displayPosn[RemoteIndex].x = 0;
  displayPosn[RemoteIndex].y = 0;
  displayPosn[EncodeIndex].y = 0;

  double halfWidth;
  double halfHeight1;
  double halfHeight2;
  if (videoPIP) {
    halfWidth  =  width[RemoteIndex] / (double)((width[EncodeIndex] >> 1) + width[RemoteIndex]);
    halfHeight1 = (height[EncodeIndex] >> 1) / (double)PMAX(height[EncodeIndex] >> 1, height[RemoteIndex]);
    halfHeight2 = height[RemoteIndex]  / (double)PMAX(height[EncodeIndex] >> 1, height[RemoteIndex]);
  } else {
    halfWidth = width[RemoteIndex] / (double)(width[RemoteIndex] + width[EncodeIndex]);
    halfHeight1 = height[EncodeIndex] / (double)PMAX(height[EncodeIndex], height[RemoteIndex]);
    halfHeight2 = height[RemoteIndex] / (double)PMAX(height[EncodeIndex], height[RemoteIndex]);
  }

  displayPosn[EncodeIndex].x = (short) (halfWidth * w);
  displayPosn[EncodeIndex].w = w - displayPosn[EncodeIndex].x;
  displayPosn[EncodeIndex].h = (short) (halfHeight1 * h);

  displayPosn[RemoteIndex].w = displayPosn[EncodeIndex].x;
  displayPosn[RemoteIndex].h = (short) (halfHeight2 * h);

  WriteOutDisplay();
}

void SdlDisplayThread::CloseScreen(void)
{ 
  if (screen != NULL) {    
    PTRACE(3,"SDL\t close the SDL display screen.");

    SDL_FreeSurface(screen);
    screen = NULL;
    oldScreenWidth = 0;
    oldScreenHeight = 0;
    SDL_Quit();
  } 
}

BOOL  SdlDisplayThread::CreateOverlay(BOOL isEncoding)
{
  unsigned dispIndex = GetDisplayIndex(isEncoding);
  unsigned w = width[dispIndex];
  unsigned h = height[dispIndex];
  
  if ((overlay[dispIndex] != NULL) && 
      ((unsigned)overlay[dispIndex]->w == w)  &&
      ((unsigned)overlay[dispIndex]->h == h)  )
   return TRUE;
  
  //AT this stage, we decide that we need a new overlay.
  //Get rid of the old overlay, then make the new overlay.    
  if (overlay[dispIndex] != NULL) 
    SDL_FreeYUVOverlay(overlay[dispIndex]);

  if ((w == 0) || (h == 0)) { //An overlay of size 0x0 is meaningless.
    overlay[dispIndex]= NULL;
    return TRUE;
  }
  
  overlay[dispIndex] = SDL_CreateYUVOverlay(w, h, SDL_IYUV_OVERLAY, screen);
 
  if (overlay[dispIndex] == NULL) {
    PTRACE(1, "SDL\t Could not open overlay to display window: " << SDL_GetError());
    return FALSE;
  }
  
  PTRACE(3, "SDL\t Successfully create the YUV overlay of size " << w << "x" << h );
  return TRUE;
}

BOOL SdlDisplayThread::SetOverlaySize (BOOL isEncoding, unsigned _width, unsigned _height)
{
  unsigned dispIndex = GetDisplayIndex(isEncoding);

  if ((width[dispIndex] != _width) || (height[dispIndex] != _height)) {
    PTRACE(3, "SDL\t Requested internal size of " << _width << "x" << _height);
    PTRACE(3, "SDL\t did not match internal size of " 
	   << width[dispIndex] << "x" << height[dispIndex]);
    
    if((_width != 0) && (_height != 0)) {
      width[dispIndex] = _width;
      height[dispIndex] = _height;
      InitDisplayPosn();
      CreateOverlay(isEncoding);
    }    
  }

  return TRUE;
}

void SdlDisplayThread::CloseWindow(BOOL isEncoding)
{
  PINDEX dispIndex = GetDisplayIndex(isEncoding);
  PTRACE(3, "SDL\tClose window " << dispIndex);

  if (!(width[dispIndex] || height[dispIndex])) {
    return;
  }

  if(overlay[dispIndex] != NULL) {
    PTRACE(3, "SDL\tClose the overlay for window " << dispIndex);
    SDL_FreeYUVOverlay(overlay[dispIndex]);
    overlay[dispIndex] = NULL;    
  }
  
  width[dispIndex] = 0;
  height[dispIndex] = 0;
  displayPosn[dispIndex].w = 0;
  displayPosn[dispIndex].h = 0;

  if(overlay[1-dispIndex] == NULL)
    CloseScreen();
  else
    ResizeScreen(displayPosn[1-dispIndex].w, displayPosn[1-dispIndex].h);
}


void SdlDisplayThread::ProcessSdlEvents(void) 
{
  if (!ScreenIsOpen()) {
    PTRACE(6, "SDL\t Screen not open, so dont process events");
    return;
  }

  SDL_Event event;  
  while (SDL_PollEvent(&event)) {
    if (event.type == SDL_QUIT) {//User selected cross
      
      PTRACE(3, "SDL\t user selected cross on window, close window");
      CloseWindow(TRUE);
      CloseWindow(FALSE);
      CloseScreen();
      displayIsShutDown = TRUE;
      return;
    }
    if (event.type == SDL_VIDEORESIZE){
      PTRACE(3, "SDL\t Resize window to " << event.resize.w << "x" << event.resize.h);
      ResizeScreen(event.resize.w, event.resize.h);
      InitDisplayPosn();
    }    //end event.type==SDL_VIDEORESIZE
  }      //end while (SDL_PollEvent())
}

BOOL SdlDisplayThread::Redraw(BOOL isEncoding, OneVideoFrame *frame)
{
  if (frame == NULL)
    return TRUE;

  unsigned dispIndex = GetDisplayIndex(isEncoding);

  PTRACE(6, "SDL\tRedraw starts now. Window " << dispIndex);  
        
  if (DisplayIsShutDown()) {   //Some task somewhere has closed our display device.
    PTRACE(6, "SDL\tScreen is closed. ::Redraw() returning immediately"); 
    return TRUE;   //(could have been user selecting cross on the window)
  }

  if ((width[dispIndex] == 0) || (height[dispIndex] == 0)) {
    unsigned w = oldScreenWidth;
    unsigned newH;
    if (isEncoding && videoPIP) {
      w += frame->GetWidth() / 2;
      newH = frame->GetHeight() / 2; 
    } else {
      w += frame->GetWidth();
      newH = frame->GetHeight();
    }
    unsigned h = PMAX(oldScreenHeight, newH);

    ResizeScreen(w, h);
  }

  SetOverlaySize(isEncoding, frame->GetWidth(), frame->GetHeight());

  SDL_Overlay *yuvOverlay= overlay[dispIndex];

  if (yuvOverlay == NULL) {
    PTRACE(6, "SDL\tRedraw end prematurely, no overlay for window " << dispIndex);
    return TRUE;       //Again, we lost the overlay, exit quickly.
  }
  
  unsigned char * base = frame->GetDataPointer();
  yuvOverlay->pixels[0] = base;
  yuvOverlay->pixels[1] = base + (width[dispIndex] * height[dispIndex]);
  yuvOverlay->pixels[2] = base + (width[dispIndex] * height[dispIndex] * 5/4);

  SDL_LockYUVOverlay(yuvOverlay);
  SDL_DisplayYUVOverlay(yuvOverlay, displayPosn + dispIndex );    
  SDL_UnlockYUVOverlay(yuvOverlay);

  PTRACE(6, "SDL\tFinish the redraw for window " << GetDirName(isEncoding));

  return TRUE;  
}

void SdlDisplayThread::WriteOutDisplay(void)
{
#ifdef PTRACING
  PINDEX i;
  for(i = 0; i < 2; i++) {
    PTRACE(6, "SDL\t" << i << " display x,y=" << displayPosn[i].x << " "
	   << displayPosn[i].y );
  
    PTRACE(6, "SDL\t" << i << " display w,h=" << displayPosn[i].w << " "
	 << displayPosn[i].h );
  
    PTRACE(6, "SDL\t" << i << " actual  w,h = " << width[i] << " " << height[i]);

    PStringStream str;
    if (overlay[i] != NULL)
      str << (void *) overlay[i] << " width, height==" << overlay[i]->w << "x" << overlay[i]->h;
    else
      str << "NULL";
    PTRACE(6, "SDL\t Window:" << i << " Overlay  is "  << str);
  }

  PTRACE(6, "SDL\t Screen is " << (void *) screen);
#endif
}

unsigned SdlDisplayThread::GetDisplayIndex(BOOL isEncoding)
{
  if (isEncoding)
    return EncodeIndex;
  else
    return RemoteIndex;
}

void SdlDisplayThread::Main()
{
  OneVideoFrame *frame;
  InitialiseSdl();

  do {
    readFlag.Wait();

    WriteOutDisplay();
    frame = GetNextFrame(TRUE);
    Redraw(TRUE, frame);
    if (frame != NULL)
      delete frame;
    
    ProcessSdlEvents();

    frame = GetNextFrame(FALSE);
    Redraw(FALSE, frame);
    if (frame != NULL)
      delete frame;
    
    ProcessSdlEvents();
    
    if (closeRecWindow) {
      CloseWindow(FALSE);     
      ProcessSdlEvents();
    }

    if (closeEncWindow) {
      CloseWindow(TRUE);
      ProcessSdlEvents();
    }

  }
  while (IsOpen());

  CloseWindow(TRUE);
  CloseWindow(FALSE);

  exitFlag.Signal();
}
///////////////////////////////////////////////////
#endif

// HAS_SDL
