/*
 * sdl.h
 *
 * SDL video interface
 *
 * Copyright (c) 1999-2000 Equivalence Pty. Ltd.
 *
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.0 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
 * the License for the specific language governing rights and limitations
 * under the License.
 *
 * The Original Code is Open H323 Library.
 *
 * The Initial Developer of the Original Code is Equivalence Pty. Ltd.
 *
 * Contributor(s): Derek J Smithies (derek@indranet.co.nz)
 *
 * $Log: sdlvid.h,v $
 * Revision 1.4  2002/04/29 03:51:55  dereks
 * Documentation tidy up. Thanks to Walter Whitlock.
 *
 * Revision 1.3  2002/04/26 03:33:32  dereks
 * Major upgrade. All calls to SDL library are now done by one thread.
 *
 * Revision 1.2  2001/05/25 01:14:44  dereks
 * Alter SetFrameSize & OpenWindo to use unsigned variables. Change type of
 * other variables to eliminate compiler warnings.
 *
 * Revision 1.1  2001/03/07 01:47:45  dereks
 * Initial release of SDL (Simple DirectMedia Layer, a cross-platform multimedia library),
 * a video library code.
 *
 *
 */
#ifdef  HAS_SDL
#ifndef _OhPhone_SDL_H
#define _OhPhone_SDL_H

#include <ptlib.h>
#include <videoio.h>

#include <SDL.h>

class OneVideoFrame : public PObject
{
  PCLASSINFO(OneVideoFrame, PObject);
 public:
  OneVideoFrame(unsigned newWidth, unsigned newHeight, u_char *data);
  OneVideoFrame(unsigned newWidth, unsigned newHeight, const void *_data);

  ~OneVideoFrame();

  unsigned GetWidth() { return width; }
  unsigned GetHeight() { return height; }

  u_char *GetDataPointer() { return data; }

 protected:
  void Initialise(unsigned newWidth, 
		  unsigned newHeight, u_char *_data);

  unsigned width;
  unsigned height;
  
  u_char *data;
};

class SdlDisplayThread : public PThread
{
  PCLASSINFO(SdlDisplayThread, PThread);

 public:
  SdlDisplayThread(BOOL _videoPIP);

  ~SdlDisplayThread();
 
  /** returns FALSE if the thread is closed, so cannot add frame.
      The frame is deleted by the thread - in all cases.*/

  BOOL AddFrame(OneVideoFrame *newFrame, BOOL isEncoding);

  BOOL IsOpen();

  virtual void Terminate();
  void AwaitTermination();
  void RequestOpenWindow(BOOL isEncoding);
  void RequestCloseWindow(BOOL isEncoding);

  void Main();
 protected:
  BOOL ScreenIsOpen();
  BOOL DisplayIsShutDown();
  void CloseWindow(BOOL isEncoding);
  

  OneVideoFrame *GetNextFrame(BOOL isEncoding);

  BOOL InitialiseSdl();
  BOOL ResizeScreen(unsigned newWidth, unsigned newHeight);
  void InitDisplayPosn();
  void InitDisplayPosn(unsigned w, unsigned h);
  void CloseScreen();
  BOOL CreateOverlay(BOOL isEncoding);
  BOOL SetOverlaySize (BOOL isEncoding, unsigned _width, unsigned _height);

  void WriteOutDisplay();

  unsigned GetDisplayIndex(BOOL isEncoding);

  /**Store the height,widths etc for this classes window
   */
    
  BOOL SetFrameSize(BOOL isEncoding, unsigned _width, unsigned _height);

  /**Handles all events that occur in the SDL window (resize and quit)
   */
  void ProcessSdlEvents(void);
  BOOL Redraw(BOOL isEncoding, OneVideoFrame *frame);

  enum { RemoteIndex = 0 };
  enum { EncodeIndex = 1 };

  const char * GetDirName(BOOL isEncoding) 
    { return (isEncoding ? "local" : "remote"); }

  SDL_Surface  *screen;
  SDL_Overlay  *overlay[2];
  SDL_Rect      displayPosn[2];
  
  unsigned   width[2];
  unsigned   height[2];
  unsigned   oldScreenWidth, oldScreenHeight;

  PString  remoteName;
  BOOL   threadOpen;
  BOOL   displayIsShutDown;
  BOOL   videoPIP;

  BOOL  closeEncWindow;
  BOOL  closeRecWindow;

  OneVideoFrame *nextEncFrame;
  OneVideoFrame *nextRcvFrame;
  
  PMutex     accessLock;  
  PSyncPoint exitFlag;
  PSyncPoint readFlag;
};

/**Display data to the SDL screen.
  */

class SDLVideoDevice : public H323VideoDevice
{
  PCLASSINFO(SDLVideoDevice, H323VideoDevice);

 public:
  /**Constructor. Does not make a window.
   */
  SDLVideoDevice(const PString & _remoteName,
		 BOOL _isEncoding, 
		 SdlDisplayThread *_sdlThread);

  /**Destructor.  Closes window if necessary, (which initializes all variables)
   */
  ~SDLVideoDevice();

  /**Synonymous with the destructor.
   */
  BOOL Close();
  
  /**Global test function to determine if this video rendering
     class is open.*/
  BOOL IsOpen();

  /**Take a YUV411 format image, render it on the existing window. 
     If the window is not present, there is no rendering.
  */
   BOOL Redraw (const void *frame);

   /**Set size of the window. Closes existing window, opens new window.
    */
   BOOL SetFrameSize (unsigned _width ,unsigned _height);

   /**Name of remote computer, which is used for window title.
    */
   virtual PString GetRemoteName() const
     { return remoteName ; }
    
   /**Name of remote computer, which is used for window title.
    */
   virtual void SetRemoteName(const PString & _remoteName)
     { remoteName = _remoteName; }

   /**Specifies required number of bitplanes in the image. Does nothing.
    */
   void ForceDepth(int d) { return;  }


  protected:
    // overrides from H323VideoDevice
    BOOL WriteLineSegment(int /*x*/, int /*y*/, unsigned /*len*/, const BYTE * /*data*/)
       { return TRUE; }

 private:
    BOOL     isEncoding;
    PString  remoteName;
    SdlDisplayThread *sdlThread;
    unsigned     width,height;
};




#endif
#endif

// End of File ///////////////////////////////////////////////////////////////
